from __future__ import annotations

from abc import ABC, abstractmethod

from pydantic import BaseModel, Extra

from ..executor import SerializableExecutionEvent
from ..lifecycle import AsyncLifecycle

_extension_classes = dict()


def extension(t: type[Extension]):
    if not issubclass(t, Extension):
        raise TypeError(f'Must extend `Extension` type: {t.__name__}')
    _extension_classes[t.__name__] = t
    return t


class Extension(ABC, AsyncLifecycle, BaseModel):
    class Config:
        extra = Extra.forbid
        underscore_attrs_are_private = True
        use_enum_values = True

    @abstractmethod
    async def update(self, e: SerializableExecutionEvent) -> None:
        pass



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict

from pydantic import BaseModel, Extra

from .lifecycle import Lifecycle

_task_classes = dict()


def task(t: type[Task]):
    if not issubclass(t, Task):
        raise TypeError(f'Must extend `Task` type: {t.__name__}')
    _task_classes[t.__name__] = t
    return t


class Task(ABC, BaseModel, Lifecycle):
    _shared_dict: Dict[str, Any] = dict()

    @property
    def shared_dict(self) -> Dict[str, Any]:
        return self._shared_dict

    class Config:
        extra = Extra.forbid
        underscore_attrs_are_private = True
        use_enum_values = True

    @abstractmethod
    def run(self) -> None:
        pass



from __future__ import annotations

from enum import Enum
from typing import Iterable, Optional

from . import EventBus, SerializableEvent, serializable_event


class Severity(Enum):
    DEBUG = 'DEBUG'
    INFO = 'INFO'
    WARNING = 'WARNING'
    ERROR = 'ERROR'
    CRITICAL = 'CRITICAL'


class SerializableLogEvent(SerializableEvent):
    @classmethod
    def event_group(cls) -> str:
        return 'SerializableLogEvent'


@serializable_event
class LogStartEvent(SerializableLogEvent):
    name: str


@serializable_event
class LogEndEvent(SerializableLogEvent):
    name: str


@serializable_event
class LogWriteEvent(SerializableLogEvent):
    name: str
    severity: Severity
    message: str


@serializable_event
class UnknownLogEvent(SerializableLogEvent):
    content: str


class LogWriter:
    __slots__ = ('b', 'n')

    def __init__(self, n: str, b: Optional[EventBus[SerializableLogEvent]]) -> None:
        self.n = n
        self.b = b
        if self.b:
            self.b.put(LogStartEvent(name=self.n))

    def write(self, m: str) -> None:
        if self.b:
            self.b.put(LogWriteEvent(name=self.n, severity=Severity.INFO, message=m))

    def writelines(self, mlist: Iterable[str]) -> None:
        for m in mlist:
            self.write(m)

    def flush(self) -> None:
        # Need to imitate file-like object for redirecting stdout
        pass

    def close(self) -> None:
        if self.b:
            self.b.put(LogEndEvent(name=self.n))



from abc import abstractmethod
from datetime import datetime

from ...executor import SerializableExecutionEvent
from ..extension import Extension


class Notification(Extension):
    async def update(self, _: SerializableExecutionEvent) -> None:
        # We don't want notifications to be spammed...
        pass

    @abstractmethod
    def send_notification(self, title: str, msg: str) -> None:
        pass

    async def on_create(self) -> None:
        self.send_notification(
            "Flowmancer Job Notification: STARTING", f"Job initiated at {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}"
        )

    async def on_success(self) -> None:
        self.send_notification(
            "Flowmancer Job Notification: SUCCESS",
            f"Job completed successfully at {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}",
        )

    async def on_failure(self) -> None:
        self.send_notification(
            "Flowmancer Job Notification: FAILURE", f"Job failed at {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}"
        )

    async def on_abort(self) -> None:
        self.send_notification(
            "Flowmancer Job Notification: ABORTED", f"Job aborted at {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}"
        )



from enum import Enum

from . import SerializableEvent, serializable_event


class SerializableExecutionEvent(SerializableEvent):
    @classmethod
    def event_group(cls) -> str:
        return 'SerializableExecutionEvent'


class ExecutionState(Enum):
    FAILED = 'F'
    PENDING = 'P'
    RUNNING = 'R'
    DEFAULTED = 'D'
    COMPLETED = 'C'
    ABORTED = 'A'
    SKIP = 'S'
    INIT = '_'


@serializable_event
class ExecutionStateTransition(SerializableExecutionEvent):
    name: str
    from_state: ExecutionState
    to_state: ExecutionState
